from .article import ArticleDisplayConfig, DependantArticleDisplayConfig
from .article_type import ArticleTypeDisplayConfig
from .in_editor_template import InEditorTemplateDisplayConfig
from .meta_information import MetaInformationDisplayConfig
from .meta_information_instance import MetaInformationInstanceDisplayConfig
from .publication import PublicationDisplayConfig, PublicationParserDisplayConfig



import json
import os
from abc import ABC
from datetime import date
from io import BytesIO
from typing import Iterable

from django.conf import settings
from django.core.files.base import ContentFile
from django.template.loader import render_to_string
from PIL import Image, ImageOps
from slugify import slugify
from wbcore.contrib.tags.models import Tag

from wbwriter.pdf_generator import PdfGenerator

from .typings import ArticleDTO


class ParserValidationException(Exception):
    def __init__(self, *args, errors=None, **kwargs):  # real signature unknown
        self.errors = errors


class PublicationParser(ABC):  # pragma: no cover
    def __init__(self, article: ArticleDTO, published_date: date):
        self.article = article
        self.published_date = published_date

    @property
    def template_path(self) -> str:
        if hasattr(self, "TEMPLATE_PATH"):
            return self.TEMPLATE_PATH
        raise ValueError(
            f"The attribute TEMPLATE_PATH needs to be define in the parser {self.__class__.__name__} in order to generate file"
        )

    @property
    def article_type(self) -> str:
        if hasattr(self, "ARTICLE_TYPE"):
            return self.ARTICLE_TYPE
        raise ValueError(
            f"The attribute ARTICLE_TYPE needs to be define in the parser {self.__class__.__name__} in order to generate content"
        )

    @property
    def default_section(self) -> str:
        if hasattr(self, "DEFAULT_SECTION"):
            return self.DEFAULT_SECTION
        raise ValueError(
            f"The attribute DEFAULT_SECTION needs to be define in the parser {self.__class__.__name__} in order to generate content"
        )

    @property
    def errors(self) -> list[str]:
        if not hasattr(self, "_errors"):
            raise AssertionError("You must call `.is_valid()` before accessing `.errors`.")
        return self._errors

    def _is_valid(self) -> bool:
        return True

    def is_valid(self, raise_exception: bool = True) -> bool:
        self._errors = []
        is_valid = self._is_valid()
        if self.errors and raise_exception:
            raise ParserValidationException(errors=self.errors)
        return is_valid

    def _get_additional_information(self) -> dict[str, any]:
        additional_information = {"type": self.article_type, "is-private": self.article.is_private}

        for meta_info in self.article.meta_information:
            additional_information[meta_info["meta_information__key"]] = meta_info["boolean_value"]
        return additional_information

    def get_file(self, context: dict[str, any] | None = None) -> BytesIO | None:
        """Parses the given object's data and returns the publications content
        rendered as a PDF.

        Parameters
        ----------
        context: The context data that should be used by the file. Optional. If not provided, it will be fetchted from get_file_context

        Returns
        -------
        Either None or ByteIO. The latter representing the PDF representation of the content.
        """
        if not context:
            context = self.get_file_context()
        html = render_to_string(self.template_path, context)

        if settings.DEBUG:
            if not os.path.isdir("tmp"):
                os.mkdir("tmp")
            with open(f"tmp/{slugify(self.article.title)}.html", "w") as f:
                f.write(html)

        pdf_generator = PdfGenerator(
            main_html=html,
            custom_css=[],
            side_margin=0,
            extra_vertical_margin=0,
        )

        return pdf_generator.render_pdf()

    def _get_images(self) -> tuple[None | tuple[str, ContentFile], None | tuple[str, ContentFile]]:
        """Derives the teaser image and its thumbnail form the given object.

        On the first position there is a tuple that contains the filename and
        the ContentFile for the teaser image, if it is not None.

        On the second position there is a tuple that contains the filename and
        the ContentFile for the thumbnail of the teaser image, if it is not
        None.

        Parameters
        ----------
        article: The data that should be published.

        Returns
        -------
        Returns a tuple that may contain the data for a teaser image and
        its thumbnail.

        On the first position there is a tuple that contains the filename and
        the ContentFile for the teaser image, if it is not None.

        On the second position there is a tuple that contains the filename and
        the ContentFile for the thumbnail of the teaser image, if it is not
        None.
        """
        teaser_image = self.article.teaser_image
        if not teaser_image:
            return (None, None)

        basename_parts = os.path.basename(teaser_image.name).rsplit(".", 1)
        if len(basename_parts) < 2:
            teaser_image_basename = basename_parts[0]
            teaser_image_extension = "png"
        else:
            teaser_image_basename = basename_parts[0]
            teaser_image_extension = basename_parts[1]

        if len(teaser_image_basename) > 100:
            teaser_image_basename = teaser_image_basename.substring(0, 92)
        if teaser_image_extension == "jpg":  # NOTE: Image doesn't understand "jpg"
            teaser_image_extension = "jpeg"
        pil_image_obj = Image.open(self.article.teaser_image)

        pil_image_obj = ImageOps.exif_transpose(
            pil_image_obj
        )  # We need transpose the image accoording to the Exif data for mobile phone shots

        pil_teaser_image_obj = pil_image_obj.copy()
        pil_teaser_image_obj.thumbnail((1000, 1000))
        teaser_image_tuple = None
        f_teaser = BytesIO()
        try:
            pil_teaser_image_obj.save(f_teaser, format=teaser_image_extension)
            teaser_image_tuple = (
                f"{teaser_image_basename}.{teaser_image_extension}",
                ContentFile(f_teaser.getvalue()),
            )
        finally:
            f_teaser.close()

        pil_thumbnail_image_obj = pil_image_obj.copy()
        pil_thumbnail_image_obj.thumbnail((510, 510))
        thumbnail_image_tuple = None
        f_thumbnail = BytesIO()
        try:
            pil_thumbnail_image_obj.save(f_thumbnail, format=teaser_image_extension)
            thumbnail_image_tuple = (
                f"{teaser_image_basename}_thumbnail.{teaser_image_extension}",
                ContentFile(f_thumbnail.getvalue()),
            )
        finally:
            f_thumbnail.close()

        if teaser_image_tuple and thumbnail_image_tuple:
            return (teaser_image_tuple, thumbnail_image_tuple)
        if teaser_image_tuple:
            return (teaser_image_tuple, None)
        if thumbnail_image_tuple:
            return (None, thumbnail_image_tuple)
        return (None, None)

    def _get_tags(self) -> Iterable[tuple[str, str]]:
        """Derives the necessary tags from the object data..

        Parameters
        ----------
        article: The data that should be published.

        Returns
        -------
        A list of key, label tag

        """
        for tag_id in self.article.tags:
            tag = Tag.objects.get(id=tag_id)
            yield (tag.slug, tag.title)

    def _get_target(self) -> str:
        """Returns a string that is unique for the publication target.

        Example:

            "website" if the publication goes to a companies own website.
        """
        return "website"

    def get_publication_content(self) -> str:
        return json.dumps({"content": self.article.content})

    def get_file_context(self) -> dict[str, str]:
        return self.article.to_dict()

    def parse(self, publication) -> None:
        """Parses the data given in `article` and changes (but doesn't save) the
        given publication instance accordingly.

        Parameters
        ----------
        publication: An instance of the Publication model. This instance will
        be changed based on the given data in `article`. The instance is not saved! You need to call the save method
        yourself.

        article: The data that should be parsed into a publication.

        published_date: The date that the publication should display as the
        publication date.
        """

        if self.is_valid():
            publication.additional_information = self._get_additional_information()

            publication.content = self.get_publication_content()

            file = self.get_file()
            if file:
                publication.content_file.save(f"{publication.slug}.pdf", BytesIO(file))

            (teaser_image, thumbnail_image) = self._get_images()
            if teaser_image and len(teaser_image) == 2:
                publication.teaser_image.save(teaser_image[0], teaser_image[1])
            if thumbnail_image and len(thumbnail_image) == 2:
                publication.thumbnail_image.save(thumbnail_image[0], thumbnail_image[1])

            publication.tags.clear()
            for tag_key, tag_title in self._get_tags():
                tag = Tag.objects.get_or_create(
                    slug=slugify(tag_key), content_type=None, defaults={"title": tag_title}
                )[0]
                publication.tags.add(tag)

            publication.target = self._get_target()



from django.db import models


class Template(models.Model):
    """
    A template consists of n styles and a template which holds a {{content}} templatetag where the content is rendered
    into.
    """

    title = models.CharField(max_length=255, unique=True)
    template = models.TextField(default="")

    header_template = models.ForeignKey(
        "Template",
        null=True,
        blank=True,
        related_name="header_templates",
        on_delete=models.SET_NULL,
    )
    footer_template = models.ForeignKey(
        "Template",
        null=True,
        blank=True,
        related_name="footer_templates",
        on_delete=models.SET_NULL,
    )

    styles = models.ManyToManyField(to="wbwriter.Style", related_name="templates", blank=True)

    side_margin = models.FloatField(default=2.5)
    extra_vertical_margin = models.FloatField(default=10)

    def __str__(self):
        return self.title



from typing import Optional

from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    Display,
    create_simple_display,
)
from wbcore.metadata.configs.display.instance_display.utils import repeat_field
from wbcore.metadata.configs.display.view_config import DisplayViewConfig


class InEditorTemplateDisplayConfig(DisplayViewConfig):
    """Provides getter methods for the list and instance displays."""

    def get_instance_display(self) -> Display:
        return create_simple_display(
            [
                ["uuid", "title"],
                ["style", "template"],
                ["modified", "is_stand_alone_template"],
                [repeat_field(2, "description")],
            ]
        )

    def get_list_display(self) -> Optional[dp.ListDisplay]:
        """Sets up and returns the list display for InEditorTemplate."""
        return dp.ListDisplay(
            fields=[
                dp.Field(key="uuid", label="UUID"),
                dp.Field(key="title", label="Title"),
                dp.Field(key="description", label="Description"),
                # dp.Field(key="style", label="Style"),
                # dp.Field(key="template", label="Template"),
                dp.Field(key="modified", label="Last modified"),
                dp.Field(key="is_stand_alone_template", label="Is a Stand-Alone Template"),
            ],
            formatting=[],
            legends=[],
        )



from wbcore.viewsets import ModelViewSet, RepresentationViewSet

from wbwriter.models.meta_information import MetaInformationInstance
from wbwriter.serializers.meta_information import (
    MetaInformationInstanceModelSerializer,
    MetaInformationInstanceRepresentationSerializer,
)

from .display import MetaInformationInstanceDisplayConfig
from .endpoints import MetaInformationInstanceEndpointConfig


class MetaInformationInstanceModelViewSet(ModelViewSet):
    serializer_class = MetaInformationInstanceModelSerializer
    queryset = MetaInformationInstance.objects.all()

    display_config_class = MetaInformationInstanceDisplayConfig
    endpoint_config_class = MetaInformationInstanceEndpointConfig

    search_fields = None
    ordering_fields = ("boolean_value",)

    def get_queryset(self):
        qs = super().get_queryset()
        if article_id := self.kwargs.get("article_id", None):
            qs = qs.filter(article_id=article_id)
        return qs.select_related("meta_information")


class MetaInformationInstanceRepresentationViewSet(RepresentationViewSet):
    serializer_class = MetaInformationInstanceRepresentationSerializer
    queryset = MetaInformationInstance.objects.all()

    filter_fields = {
        # "article": ["exact"],
        "meta_information": ["exact"],
        "boolean_value": ["exact"],
    }
    ordering_fields = ("meta_information__name", "boolean_value")

    def get_queryset(self):
        request = self.request
        user = request.user
        profile = user.profile

        if user.is_superuser or profile.is_internal:
            return MetaInformationInstance.objects.all()

        return MetaInformationInstance.objects.none()



from rest_framework.reverse import reverse
from wbcore.metadata.configs.endpoints import EndpointViewConfig


class MetaInformationInstanceEndpointConfig(EndpointViewConfig):
    def get_create_endpoint(self, **kwargs):
        return None

    def get_endpoint(self, **kwargs):
        if article_id := self.view.kwargs.get("article_id", None):
            return reverse(
                "wbwriter:metainformationinstancearticle-list", kwargs={"article_id": article_id}, request=self.request
            )
        return reverse("wbwriter:metainformationinstance-list", request=self.request)



from wbcore.viewsets import ModelViewSet, RepresentationViewSet

from wbwriter.models import MetaInformation
from wbwriter.serializers import MetaInformationModelSerializer
from wbwriter.serializers.meta_information import (
    MetaInformationRepresentationSerializer,
)

from .display import MetaInformationDisplayConfig


class MetaInformationRepresentationViewSet(RepresentationViewSet):
    serializer_class = MetaInformationRepresentationSerializer
    queryset = MetaInformation.objects.all()

    search_fields = ("name", "key")
    filter_fields = {
        "article_type": ["exact"],
        "name": ["icontains", "exact"],
        "key": ["icontains", "exact"],
        "meta_information_type": ["exact"],
    }
    ordering_fields = ("article_type", "name", "key", "meta_information_type")

    def get_queryset(self):
        request = self.request
        user = request.user
        profile = user.profile

        if user.is_superuser or profile.is_internal:
            return MetaInformation.objects.all()

        return MetaInformation.objects.none()


class MetaInformationModelViewSet(ModelViewSet):
    serializer_class = MetaInformationModelSerializer
    queryset = MetaInformation.objects.all()

    display_config_class = MetaInformationDisplayConfig

    search_fields = (
        "name",
        "key",
    )
    filterset_fields = {
        "article_type": ["exact"],
        "name": ["icontains"],
        "key": ["icontains"],
        "meta_information_type": ["exact"],
    }
    ordering_fields = ("article_type", "name", "key", "meta_information_type")



from typing import Optional

from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    Display,
    create_simple_display,
)
from wbcore.metadata.configs.display.view_config import DisplayViewConfig


class ArticleTypeDisplayConfig(DisplayViewConfig):
    """Provides getter methods for the list and instance displays."""

    def get_instance_display(self) -> Display:
        return create_simple_display([["label"]])

    def get_list_display(self) -> Optional[dp.ListDisplay]:
        """Sets up and returns the list display for ArticleType."""
        return dp.ListDisplay(
            fields=[
                dp.Field(key="id", label="ID"),
                dp.Field(key="label", label="Label"),
                dp.Field(key="slug", label="Slug"),
            ],
            formatting=[],
            legends=[],
        )



from typing import Optional

from django.utils.translation import gettext as _
from wbcore.contrib.color.enums import WBColor
from wbcore.contrib.icons import WBIcon
from wbcore.enums import Operator, Unit
from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.instance_display import (
    Display,
    Inline,
    Layout,
    Page,
    Section,
    Style,
)
from wbcore.metadata.configs.display.instance_display.operators import default
from wbcore.metadata.configs.display.view_config import DisplayViewConfig

from wbwriter.models import Article


class DependantArticleDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> dp.ListDisplay:
        article = "dependant_article" if "article_id" in self.view.kwargs else "article"
        return dp.ListDisplay(
            fields=[
                dp.Field(key=article, label=_("Article")),
            ]
        )


class ArticleDisplayConfig(DisplayViewConfig):
    """Provides getter methods for the list and instance displays."""

    def get_instance_display(self) -> Display:
        using_section = Section(
            key="using_section",
            title=_("Uses Articles"),
            collapsible=False,
            display=Display(
                pages=[
                    Page(
                        layouts={
                            default(): Layout(
                                grid_template_areas=[["dependantarticle-article"]],
                                inlines=[
                                    Inline(key="dependantarticle-article", endpoint="dependantarticle-article"),
                                ],
                            )
                        },
                    )
                ]
            ),
        )

        used_section = Section(
            key="used_section",
            title=_("Used in Articles"),
            collapsible=False,
            display=Display(
                pages=[
                    Page(
                        layouts={
                            default(): Layout(
                                grid_template_areas=[["usedarticle-article"]],
                                inlines=[
                                    Inline(key="usedarticle-article", endpoint="usedarticle-article"),
                                ],
                            )
                        },
                    )
                ]
            ),
        )

        status_section = Section(
            key="status_section",
            title=_("Status"),
            collapsible=False,
            display=Display(
                pages=[
                    Page(
                        title=_("Status"),
                        layouts={
                            default(): Layout(
                                grid_template_areas=[["status"]],
                                grid_template_columns="auto-fill",
                            )
                        },
                    )
                ]
            ),
        )

        permission_section = Section(
            key="permission_section",
            title=_("Author & Reviewers"),
            collapsed=True,
            display=Display(
                pages=[
                    Page(
                        title=_("Author & Reviewers"),
                        layouts={
                            default(): Layout(
                                grid_template_areas=[["author", "reviewer"], ["peer_reviewer", "qa_reviewer"]],
                                grid_template_columns=[Style.fr(1), Style.fr(1)],
                                grid_auto_rows=Style.MIN_CONTENT,
                            )
                        },
                    )
                ]
            ),
        )

        additional_information_section = Section(
            key="additional_information_section",
            title=_("Additional Information"),
            display=Display(
                pages=[
                    Page(
                        layouts={
                            default(): Layout(
                                grid_template_areas=[["information"]],
                                inlines=[Inline(key="information", endpoint="metainformationinstance")],
                            )
                        }
                    )
                ]
            ),
        )

        def _get_instance_display(
            collapsed_metadata_section: bool = False,
            collapsed_content_section: bool = False,
            show_is_private: bool = False,
            show_with_additional_information: bool = False,
        ) -> Display:
            """Returns a display for the wbwriter instance

            Args:
                collapsed_metadata_section: Indicates whether the metadata section should be collapsed. Defaults to False.
                collapsed_content_section: Indicates whether the content section should be collapsed. Defaults to False.
                show_is_private: Indicates whether the is_private field should be visible. Defaults to False.
                show_with_additional_information: Indicates whether the additional_information section should be added to the display. Defaults to False.

            Returns:
                Display: A display instance
            """

            def _metadata_section(collapsed: bool = True, show_is_private: bool = False) -> Section:
                """Returns the metadata section

                Args:
                    collapsed: Indicates whether the section should be collapsed or not. Defaults to True.
                    show_is_private: Indicates whether the is_private field should be visible or not. Defaults to False.

                Returns:
                    Section: A section instance
                """

                grid_template = [
                    ["teaser_image", "type", "is_private"] if show_is_private else ["teaser_image", "type", "."],
                    ["teaser_image", "name", "title"],
                    ["teaser_image", "modified", "created"],
                ]
                return Section(
                    key="metadata_section",
                    title=_("Metadata"),
                    collapsible=True,
                    collapsed=collapsed,
                    display=Display(
                        pages=[
                            Page(
                                title=_("Metadata"),
                                layouts={
                                    default(): Layout(
                                        grid_template_areas=grid_template,
                                        grid_template_columns=[Style.MIN_CONTENT, Style.fr(1), Style.fr(1)],
                                        grid_auto_rows=Style.MIN_CONTENT,
                                    ),
                                },
                            )
                        ]
                    ),
                )

            def _content_section(collapsed: bool) -> Section:
                """Returns the content section

                Args:
                    collapsed: Indicates whether the section should be collapsed or not. Defaults to False.

                Returns:
                    Section: A section instance
                """

                return Section(
                    key="content_section",
                    title=_("Content & Tags"),
                    collapsed=collapsed,
                    display=Display(
                        pages=[
                            Page(
                                title=_("Content & Tags"),
                                layouts={
                                    default(): Layout(
                                        grid_template_areas=[["tags"], ["content"]],
                                        grid_template_columns=[Style.fr(1)],
                                        grid_auto_rows=Style.MIN_CONTENT,
                                    ),
                                },
                            )
                        ]
                    ),
                )

            sections = [
                status_section,
                _metadata_section(collapsed_metadata_section, show_is_private),
                _content_section(collapsed_content_section),
                permission_section,
                used_section,
                using_section,
            ]
            grid_template = [
                ["status_section", "status_section"],
                ["metadata_section", "metadata_section"],
                ["content_section", "content_section"],
                ["permission_section", "permission_section"],
                ["using_section", "used_section"],
            ]

            if show_with_additional_information:
                sections.append(additional_information_section)
                grid_template.append(["additional_information_section", "additional_information_section"])

            return Display(
                pages=[
                    Page(
                        title=_("General Information"),
                        layouts={
                            default(): Layout(
                                grid_template_areas=grid_template,
                                grid_auto_rows=Style.MIN_CONTENT,
                                sections=sections,
                            ),
                        },
                    ),
                ]
            )

        if self.view.kwargs.get("pk"):
            obj: Article = self.view.get_object()
            if obj.status in [Article.Status.APPROVED, Article.Status.PUBLISHED]:
                if self.request.user.has_perm("wbwriter.administrate_article"):
                    return _get_instance_display(
                        collapsed_content_section=True, show_is_private=True, show_with_additional_information=True
                    )
                return _get_instance_display(collapsed_content_section=True)
            return _get_instance_display(collapsed_metadata_section=True)
        return _get_instance_display()

    def get_list_display(self) -> Optional[dp.ListDisplay]:
        """Sets up and returns the list display for Article."""
        return dp.ListDisplay(
            fields=[
                dp.Field(key="is_private_icon", label=" ", width=Unit.PIXEL(40)),
                dp.Field(key="id", label=_("ID")),
                dp.Field(key="title", label=_("Title")),
                dp.Field(key="teaser_image", label=_("Teaser image")),
                dp.Field(key="type", label=_("Type")),
                dp.Field(key="modified", label=_("Last Modification Date")),
                dp.Field(key="created", label=_("Creation Date")),
                dp.Field(key="tags", label=_("Tags")),
                dp.Field(key="author", label=_("Author")),
                dp.Field(key="reviewer", label=_("Reviewer")),
                dp.Field(key="peer_reviewer", label=_("Peer Reviewer")),
                dp.Field(key="qa_reviewer", label=_("QA Reviewer")),
            ],
            formatting=[
                dp.Formatting(
                    formatting_rules=[
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.YELLOW.value},
                            condition=dp.Condition(
                                operator=Operator.EQUAL,
                                value=Article.Status.DRAFT,
                            ),
                        ),
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.RED_DARK.value},
                            condition=dp.Condition(
                                operator=Operator.EQUAL,
                                value=Article.Status.FEEDBACK,
                            ),
                        ),
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.BLUE_LIGHT.value},
                            condition=dp.Condition(
                                operator=Operator.EQUAL,
                                value=Article.Status.PEER_REVIEW,
                            ),
                        ),
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.BLUE.value},
                            condition=dp.Condition(
                                operator=Operator.EQUAL,
                                value=Article.Status.QA_REVIEW,
                            ),
                        ),
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.GREEN_LIGHT.value},
                            condition=dp.Condition(
                                operator=Operator.EQUAL,
                                value=Article.Status.AUTHOR_APPROVAL,
                            ),
                        ),
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.GREEN.value},
                            condition=dp.Condition(
                                operator=Operator.EQUAL,
                                value=Article.Status.APPROVED,
                            ),
                        ),
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.GREEN_DARK.value},
                            condition=dp.Condition(
                                operator=Operator.EQUAL,
                                value=Article.Status.PUBLISHED,
                            ),
                        ),
                    ],
                    column="status",
                )
            ],
            legends=[
                dp.Legend(
                    key="status",
                    items=[
                        dp.LegendItem(
                            value=Article.Status.FEEDBACK.value,
                            label=Article.Status.FEEDBACK.label,
                            icon=WBColor.RED_DARK.value,
                        ),
                        dp.LegendItem(
                            value=Article.Status.DRAFT.value,
                            label=Article.Status.DRAFT.label,
                            icon=WBColor.YELLOW.value,
                        ),
                        dp.LegendItem(
                            value=Article.Status.PEER_REVIEW.value,
                            label=Article.Status.PEER_REVIEW.label,
                            icon=WBColor.BLUE_LIGHT.value,
                        ),
                        dp.LegendItem(
                            value=Article.Status.QA_REVIEW.value,
                            label=Article.Status.QA_REVIEW.label,
                            icon=WBColor.BLUE.value,
                        ),
                        dp.LegendItem(
                            value=Article.Status.AUTHOR_APPROVAL.value,
                            label=Article.Status.AUTHOR_APPROVAL.label,
                            icon=WBColor.GREEN_LIGHT.value,
                        ),
                        dp.LegendItem(
                            value=Article.Status.APPROVED.value,
                            label=Article.Status.APPROVED.label,
                            icon=WBColor.GREEN.value,
                        ),
                        dp.LegendItem(
                            value=Article.Status.PUBLISHED.value,
                            label=Article.Status.PUBLISHED.label,
                            icon=WBColor.GREEN_DARK.value,
                        ),
                    ],
                ),
                dp.Legend(
                    key="is_private",
                    items=[
                        dp.LegendItem(
                            value=False,
                            label=_("Public"),
                            icon=WBIcon.VIEW.icon,
                        ),
                        dp.LegendItem(
                            value=True,
                            label=_("Private"),
                            icon=WBIcon.IGNORE.icon,
                        ),
                    ],
                ),
            ],
        )



import pytest
from wbwriter.templatetags.writer import cite_article, table_of_contents


@pytest.mark.django_db
class TestWriter:
    def test_table_of_contents(self, article_factory):
        article_dict = {"article": article_factory(title="Test Article Title", name="Test Article Name")}
        toc = table_of_contents(article_dict, "ToC ")
        test_toc = {
            "title": "Test Article Title",
            "custom_enumeration": True,
            "anchor": "test-article-name",
            "enumerator": None,
            "level": 0,
            "articles": [],
        }
        assert toc == test_toc

    def test_table_of_contents_no_context(self):
        no_article_dict = {"blog": "Some Blog"}
        toc = table_of_contents(no_article_dict, "ToC ")
        test_toc = {"not_working": "<i>Table of content is only rendered in the exported PDF Version.</i>"}
        assert toc == test_toc

    def test_cite_article(self, article_factory):
        article = article_factory(title="Test Article Title", name="Test Article Name")
        article_dict: dict = {"article": article}
        ca = cite_article(article_dict, article.id)
        test_ca = {"anchor": "test-article-name", "text": "(see  Test Article Title)"}
        assert ca == test_ca

    def test_cite_article_no_context(self, article_factory):
        article = article_factory()
        no_article_dict = {"blog": "Some Blog"}
        ca = cite_article(no_article_dict, article.id)
        test_ca = {"not_working": "(article citations only work in a rendered pdf.)"}
        assert ca == test_ca



from reversion.views import RevisionMixin
from wbcore.viewsets import ModelViewSet, RepresentationViewSet

from wbwriter.models import Publication
from wbwriter.models.publication_models import PublicationParser
from wbwriter.serializers import (
    PublicationModelSerializer,
    PublicationParserRepresentationSerializer,
    PublicationParserSerializer,
    PublicationRepresentationSerializer,
)
from wbwriter.viewsets.buttons import PublicationButtonConfig
from wbwriter.viewsets.display import (
    PublicationDisplayConfig,
    PublicationParserDisplayConfig,
)
from wbwriter.viewsets.titles import PublicationTitleConfig


class PublicationModelViewSet(RevisionMixin, ModelViewSet):
    """Displays all fields of the Publication model."""

    serializer_class = PublicationModelSerializer
    queryset = Publication.objects.all()

    button_config_class = PublicationButtonConfig
    display_config_class = PublicationDisplayConfig
    title_config_class = PublicationTitleConfig

    search_fields = ("title", "author__computed_str", "slug", "content")
    filterset_fields = {
        "title": ["icontains", "exact"],
        "target": ["icontains", "exact"],
        "author": ["exact"],
        "created": ["exact", "lte", "gte"],
        "modified": ["exact", "lte", "gte"],
    }
    ordering_fields = ("title", "author", "created", "modified")

    def get_queryset(self):
        if self.request.user.has_perm("wbwriter.view_publication"):
            qs = Publication.objects.all()

            if content_type := self.request.GET.get("content_type"):
                qs = qs.filter(content_type_id=content_type)

            if object_id := self.request.GET.get("object_id"):
                qs = qs.filter(object_id=object_id)

            return qs.prefetch_related("tags").select_related("parser", "author")

        return Publication.objects.none()


class PublicationParserModelViewSet(ModelViewSet):
    display_config_class = PublicationParserDisplayConfig

    search_fields = ("title", "slug")
    filterset_fields = {
        "title": ["icontains", "exact"],
    }
    ordering_fields = ("title",)

    serializer_class = PublicationParserSerializer
    queryset = PublicationParser.objects.all()

    def get_queryset(self):
        if self.request.user.is_superuser or self.request.user.profile.is_internal:
            return PublicationParser.objects.all()

        return PublicationParser.objects.none()


class PublicationRepresentationViewSet(RepresentationViewSet):
    """Displays the title and the created date of the Publication model."""

    serializer_class = PublicationRepresentationSerializer
    queryset = Publication.objects.all()

    display_config_class = PublicationDisplayConfig
    button_config_class = PublicationButtonConfig

    search_fields = ("title", "author", "slug", "content")
    filterset_fields = {
        "title": ["icontains", "exact"],
        "target": ["icontains", "exact"],
        "author": ["exact"],
        "created": ["exact", "lte", "gte"],
        "modified": ["exact", "lte", "gte"],
    }
    ordering_fields = ("title", "author", "created", "modified")

    def get_queryset(self):
        if self.request.user.is_superuser or ((person := self.request.user.profile) and person.is_internal):
            return Publication.objects.all()

        return Publication.objects.none()


class PublicationParserRepresentationViewSet(RepresentationViewSet):
    """Displays the title of the PublicationParser model."""

    serializer_class = PublicationParserRepresentationSerializer
    queryset = PublicationParser.objects.all()

    display_config_class = PublicationParserDisplayConfig

    search_fields = ("title", "slug")
    filterset_fields = {
        "title": ["icontains", "exact"],
    }
    ordering_fields = ("title",)

    def get_queryset(self):
        if self.request.user.is_superuser or self.request.user.profile.is_internal:
            return PublicationParser.objects.all()

        return PublicationParser.objects.none()



from wbcore.serializers import ModelSerializer, RepresentationSerializer

from wbwriter.models import MetaInformation, MetaInformationInstance

from .article_type import ArticleTypeRepresentationSerializer


class MetaInformationModelSerializer(ModelSerializer):
    """Serializes all fields of the MetaInformation model."""

    _article_type = ArticleTypeRepresentationSerializer(source="article_type", many=True)

    class Meta:
        model = MetaInformation
        fields = (
            "id",
            "article_type",
            "_article_type",
            "name",
            "key",
            "meta_information_type",
            "boolean_default",
        )


class MetaInformationRepresentationSerializer(RepresentationSerializer):
    """Serializes the name and the key of the MetaInformation model."""

    class Meta:
        model = MetaInformation
        fields = ("id", "name", "key")


class MetaInformationInstanceModelSerializer(ModelSerializer):
    # from wbwriter.serializers import ArticleFullModelSerializer

    # _article = ArticleFullModelSerializer(source="article")
    _meta_information = MetaInformationRepresentationSerializer(source="meta_information")

    class Meta:
        model = MetaInformationInstance
        read_only_fields = ("meta_information",)
        fields = (
            "id",
            "meta_information",
            "_meta_information",
            # "article",
            # "_article",
            "boolean_value",
        )


class MetaInformationInstanceRepresentationSerializer(RepresentationSerializer):
    # from wbwriter.serializers import ArticleFullModelSerializer

    # _article = ArticleFullModelSerializer(source="article")
    _meta_information = MetaInformationRepresentationSerializer(source="meta_information")

    class Meta:
        model = MetaInformationInstance
        fields = (
            "id",
            "meta_information",
            "_meta_information",
            # "article",
            # "_article",
            "boolean_value",
        )



import pytest
from wbcore.test import GenerateTest, default_config

config = {}

for key, value in default_config.items():
    config[key] = list(
        filter(
            lambda x: x.__module__.startswith("wbwriter"),
            value,
        )
    )


@pytest.mark.django_db
@GenerateTest(config)
class TestProject:
    pass



import pytest
from rest_framework.test import APIRequestFactory
from wbcore.contrib.authentication.factories import InternalUserFactory
from wbwriter.factories import ArticleFactory, MetaInformationInstanceFactory
from wbwriter.filters import MetaInformationInstanceFilter
from wbwriter.models import MetaInformationInstance
from wbwriter.viewsets import ArticleModelViewSet


@pytest.mark.django_db
class TestArticleFilters:
    def test_get_my_articles_true(self):
        employee_a = InternalUserFactory().profile
        employee_b = InternalUserFactory().profile
        user_a = employee_a.user_account
        ArticleFactory(author=user_a.profile)
        ArticleFactory(author=employee_b)
        ArticleFactory(author=employee_b, reviewer=user_a.profile)

        request = APIRequestFactory().get("")
        request.user = user_a
        mvs = ArticleModelViewSet(kwargs={})
        qs = mvs.get_serializer_class().Meta.model.objects.all()
        filtered_qs = mvs.filterset_class(request=request).get_my_articles(qs, "", True)
        assert filtered_qs.count() == 2
        assert filtered_qs.count() < qs.count()

    def test_get_my_articles_false(self):
        employee_a = InternalUserFactory.create().profile
        employee_b = InternalUserFactory.create().profile
        user_a = employee_a.user_account
        ArticleFactory(author=user_a.profile)
        ArticleFactory(author=employee_b)
        ArticleFactory(author=employee_b, reviewer=user_a.profile)

        request = APIRequestFactory().get("")
        request.user = user_a
        mvs = ArticleModelViewSet(kwargs={})
        qs = mvs.get_serializer_class().Meta.model.objects.all()
        filtered_qs = mvs.filterset_class(request=request).get_my_articles(qs, "", False)
        assert filtered_qs.count() == qs.count()

    def test_filter_for_article_with_id(self):
        article = ArticleFactory()
        MetaInformationInstanceFactory(article=article)
        MetaInformationInstanceFactory.create_batch(4)
        qs = MetaInformationInstance.objects.all()
        filtered_qs = MetaInformationInstanceFilter.filter_for_article(self, queryset=qs, name="", value=article.id)
        assert filtered_qs.count() == 1
        assert filtered_qs.count() < qs.count()

    def test_filter_for_article_no_id(self):
        article = ArticleFactory()
        MetaInformationInstanceFactory(article=article)
        MetaInformationInstanceFactory.create_batch(4)
        qs = MetaInformationInstance.objects.all()
        filtered_qs = MetaInformationInstanceFilter.filter_for_article(self, queryset=qs, name="", value=None)
        assert filtered_qs.count() == qs.count()



from django.db import models
from django.dispatch import receiver
from wbcore.models import WBModel


class MetaInformation(WBModel):
    class MetaInformationType(models.TextChoices):
        # NOTE: This is just one value for now. However, this structure allows us to add more data type in the future.
        BOOLEAN = "BOOLEAN", "Boolean"

    article_type = models.ManyToManyField(to="wbwriter.ArticleType", related_name="meta_information")
    name = models.CharField(max_length=255, null=False, blank=False, unique=True)
    key = models.CharField(max_length=255, null=False, blank=False, unique=True)
    meta_information_type = models.CharField(
        max_length=24, choices=MetaInformationType.choices, default=MetaInformationType.BOOLEAN
    )
    boolean_default = models.BooleanField(null=True, blank=True)

    def __str__(self) -> str:
        return f"{self.name} ({self.key})"

    class Meta:
        verbose_name = "Meta Information"
        verbose_name_plural = "Meta Information"

    @classmethod
    def get_endpoint_basename(self):
        return "wbwriter:metainformation"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbwriter:metainformation-list"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{name}} ({{key}})"


class MetaInformationInstance(WBModel):
    article = models.ForeignKey(to="wbwriter.Article", related_name="meta_information", on_delete=models.CASCADE)

    meta_information = models.ForeignKey(
        to="wbwriter.MetaInformation", related_name="instances", on_delete=models.CASCADE
    )
    boolean_value = models.BooleanField(null=True, blank=True)

    def __str__(self) -> str:
        return f"{self.meta_information} / {self.article}: {self.boolean_value}"

    class Meta:
        verbose_name = "Meta Information Instance"
        verbose_name_plural = "Meta Information Instances"

        constraints = [
            models.UniqueConstraint(fields=["meta_information", "article"], name="unique_meta_information_article")
        ]

    @classmethod
    def get_endpoint_basename(self):
        return "wbwriter:metainformationinstance"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbwriter:metainformationinstance-list"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{meta_information.name}}"


@receiver(models.signals.post_save, sender="wbwriter.Article")
def create_meta_information_instances(sender, instance, created, **kwargs):
    if created and instance.type:
        for meta_information in instance.type.meta_information.all():
            MetaInformationInstance.objects.get_or_create(
                meta_information=meta_information,
                article=instance,
                defaults={"boolean_value": meta_information.boolean_default},
            )



# Generated by Django 5.0.12 on 2025-03-03 10:28

import modeltrans.fields
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('wbwriter', '0011_auto_20240103_0953'),
    ]

    operations = [
        migrations.AddField(
            model_name='article',
            name='i18n',
            field=modeltrans.fields.TranslationField(fields=['title', 'content', 'slug'], required_languages=(), virtual_fields=True),
        ),
        migrations.AddField(
            model_name='publication',
            name='i18n',
            field=modeltrans.fields.TranslationField(fields=['title', 'slug', 'content'], required_languages=(), virtual_fields=True),
        ),
    ]



from django.urls import include, path
from wbcore.routers import WBCoreRouter

from wbwriter import viewsets

router = WBCoreRouter()
router.register(r"publication", viewsets.PublicationModelViewSet, basename="publication")
router.register(
    r"publicationrepresentation", viewsets.PublicationRepresentationViewSet, basename="publicationrepresentation"
)
router.register(
    r"publicationparserrepresentation",
    viewsets.PublicationParserRepresentationViewSet,
    basename="publicationparserrepresentation",
)
router.register(
    r"publicationparser",
    viewsets.PublicationParserModelViewSet,
    basename="publicationparser",
)

router.register(r"article", viewsets.ArticleModelViewSet, basename="article")
router.register(r"articlerepresentation", viewsets.ArticleRepresentionViewSet, basename="articlerepresentation")
# router.register(r"ddq", viewsets.DDQArticleModelViewSet, basename="ddq")
router.register(r"review-article", viewsets.ReviewerArticleModelViewSet, basename="review-article")
router.register(
    r"articletyperepresentation",
    viewsets.ArticleTypeRepresentationViewSet,
    basename="articletyperepresentation",
)
router.register(
    r"articletype",
    viewsets.ArticleTypeModelViewSet,
    basename="articletype",
)
router.register(
    r"in-editor-template",
    viewsets.InEditorTemplateModelViewSet,
    basename="in-editor-template",
)
router.register(
    r"in-editor-template-representation",
    viewsets.InEditorTemplateRepresentationViewSet,
    basename="in-editor-template-representation",
)
router.register(
    r"metainformation",
    viewsets.MetaInformationRepresentationViewSet,
    basename="metainformation",
)
router.register(
    r"metainformationinstance",
    viewsets.MetaInformationInstanceModelViewSet,
    basename="metainformationinstance",
)
router.register(
    r"dependantarticle",
    viewsets.DependantArticleModelViewSet,
    basename="dependantarticle",
)
article_router = WBCoreRouter()
article_router.register(
    r"metainformationinstancearticle",
    viewsets.MetaInformationInstanceModelViewSet,
    basename="metainformationinstancearticle",
)
article_router.register(
    r"dependantarticle",
    viewsets.DependantArticleModelViewSet,
    basename="dependantarticle-article",
)

used_article_router = WBCoreRouter()
used_article_router.register(
    r"usedarticle",
    viewsets.DependantArticleModelViewSet,
    basename="usedarticle-article",
)

urlpatterns = [
    path("", include(router.urls)),
    path("article/<int:article_id>/", include(article_router.urls)),
    path("article/<int:used_article_id>/", include(used_article_router.urls)),
]



from wbcore.menus import ItemPermission, Menu, MenuItem
from wbcore.permissions.shortcuts import is_internal_user

WRITER_MENU = Menu(
    label="Writer",
    items=[
        MenuItem(
            label="Article",
            endpoint="wbwriter:article-list",
            add=MenuItem(
                label="Create an article",
                endpoint="wbwriter:article-list",
            ),
            permission=ItemPermission(
                method=lambda request: is_internal_user(request.user), permissions=["wbwriter.view_article"]
            ),
        ),
        # MenuItem(
        #     label="DDQ",
        #     endpoint="wbwriter:ddq-list",
        #     add=MenuItem(
        #         label="Create a DDQ related entry",
        #         endpoint="wbwriter:ddq-list",
        #     ),
        #     permission=ItemPermission(permissions=["wbwriter.view_article"]),
        # ),
        MenuItem(
            label="Articles to review",
            endpoint="wbwriter:review-article-list",
            permission=ItemPermission(
                method=lambda request: is_internal_user(request.user), permissions=["wbwriter.view_article"]
            ),
        ),
        MenuItem(
            label="Publication",
            endpoint="wbwriter:publication-list",
            permission=ItemPermission(
                method=lambda request: is_internal_user(request.user), permissions=["wbwriter.view_publication"]
            ),
        ),
    ],
)



from django.db import models


class PublishableMixin(models.Model):
    def get_publication_metadata(self) -> dict[str, str]:
        raise NotImplementedError("You must implement get_publication_metadata")

    class Meta:
        abstract = True



import base64

from weasyprint import CSS, HTML
from weasyprint.text.fonts import FontConfiguration


class PdfGenerator:
    """
    Generate a PDF out of a rendered template, with the possibility to integrate nicely
    a header and a footer if provided.

    Notes:
    ------
    - When Weasyprint renders an html into a PDF, it goes though several intermediate steps.
      Here, in this class, we deal mostly with a box representation: 1 `Document` have 1 `Page`
      or more, each `Page` 1 `Box` or more. Each box can contain other box. Hence the recursive
      method `get_element` for example.
      For more, see:
      https://weasyprint.readthedocs.io/en/stable/hacking.html#dive-into-the-source
      https://weasyprint.readthedocs.io/en/stable/hacking.html#formatting-structure
    - Warning: the logic of this class relies heavily on the internal Weasyprint API. This
      snippet was written at the time of the release 47, it might break in the future.
    - This generator draws its inspiration and, also a bit of its implementation, from this
      discussion in the library github issues: https://github.com/Kozea/WeasyPrint/issues/92
    """

    OVERLAY_LAYOUT = "@page {size: A4 portrait; margin: 0;}"

    def __init__(
        self,
        main_html,
        header_html=None,
        footer_html=None,
        custom_css=[],
        base_url=None,
        side_margin=2,
        extra_vertical_margin=30,
    ):
        """
        Parameters
        ----------
        main_html: str
            An HTML file (most of the time a template rendered into a string) which represents
            the core of the PDF to generate.
        header_html: str
            An optional header html.
        footer_html: str
            An optional footer html.
        custom_css: list
            An optional list of css objects. (see weasyprint.CSS)
        base_url: str
            An absolute url to the page which serves as a reference to Weasyprint to fetch assets,
            required to get our media.
        side_margin: int, interpreted in cm, by default 2cm
            The margin to apply on the core of the rendered PDF (i.e. main_html).
        extra_vertical_margin: int, interpreted in pixel, by default 30 pixels
            An extra margin to apply between the main content and header and the footer.
            The goal is to avoid having the content of `main_html` touching the header or the
            footer.
        """
        self.main_html = main_html
        self.header_html = header_html
        self.footer_html = footer_html
        self.custom_css = custom_css
        self.base_url = base_url
        self.side_margin = side_margin
        self.extra_vertical_margin = extra_vertical_margin

    def _compute_overlay_element(self, element: str):
        """
        Parameters
        ----------
        element: str
            Either 'header' or 'footer'

        Returns
        -------
        element_body: BlockBox
            A Weasyprint pre-rendered representation of an html element
        element_height: float
            The height of this element, which will be then translated in a html height
        """
        html = HTML(
            string=getattr(self, f"{element}_html"),
            base_url=self.base_url,
        )
        element_doc = html.render(stylesheets=[CSS(string=self.OVERLAY_LAYOUT)] + self.custom_css)
        element_page = element_doc.pages[0]
        element_body = PdfGenerator.get_element(element_page._page_box.all_children(), "body")
        element_body = element_body.copy_with_children(element_body.all_children())
        element_html = PdfGenerator.get_element(element_page._page_box.all_children(), element)

        if element == "header":
            element_height = element_html.height
        if element == "footer":
            element_height = element_page.height - element_html.position_y

        return element_body, element_height

    def _apply_overlay_on_main(self, main_doc, header_body=None, footer_body=None):
        """
        Insert the header and the footer in the main document.

        Parameters
        ----------
        main_doc: Document
            The top level representation for a PDF page in Weasyprint.
        header_body: BlockBox
            A representation for an html element in Weasyprint.
        footer_body: BlockBox
            A representation for an html element in Weasyprint.
        """
        for page in main_doc.pages:
            page_body = PdfGenerator.get_element(page._page_box.all_children(), "body")

            if header_body:
                page_body.children += header_body.all_children()
            if footer_body:
                page_body.children += footer_body.all_children()

    def render_pdf(self):
        """
        Returns
        -------
        pdf: a bytes sequence
            The rendered PDF.
        """
        if self.header_html:
            header_body, header_height = self._compute_overlay_element("header")
        else:
            header_body, header_height = None, 0
        if self.footer_html:
            footer_body, footer_height = self._compute_overlay_element("footer")
        else:
            footer_body, footer_height = None, 0

        margins = "{header_size}px {side_margin} {footer_size}px {side_margin}".format(
            header_size=header_height + self.extra_vertical_margin,
            footer_size=footer_height + self.extra_vertical_margin,
            side_margin=f"{self.side_margin}cm",
        )
        content_print_layout = "@page {size: A4 portrait; margin: %s;}" % margins

        html = HTML(
            string=self.main_html,
            base_url=self.base_url,
        )
        font_config = FontConfiguration()
        main_doc = html.render(
            stylesheets=[CSS(string=content_print_layout)] + self.custom_css,
            font_config=font_config,
        )

        if self.header_html or self.footer_html:
            self._apply_overlay_on_main(main_doc, header_body, footer_body)
        pdf = main_doc.write_pdf()
        encoded = base64.b64encode(pdf)
        encoded = encoded.decode("utf-8")
        return pdf

    @staticmethod
    def get_element(boxes, element):
        """
        Given a set of boxes representing the elements of a PDF page in a DOM-like way, find the
        box which is named `element`.

        Look at the notes of the class for more details on Weasyprint insides.
        """
        for box in boxes:
            if box.element_tag == element:
                return box
            return PdfGenerator.get_element(box.all_children(), element)



import json

from django.db import models


class BlockParameter(models.Model):
    block = models.ForeignKey(to="wbwriter.Block", related_name="parameters", on_delete=models.CASCADE)

    order = models.PositiveIntegerField()
    title = models.CharField(max_length=255)
    many = models.BooleanField(default=False)

    def __str__(self):
        return f"{self.block.title}: {self.title}"

    def parse_parameter(self, parameter):
        if self.many:
            return json.loads(parameter)
        return parameter



# Generated by Django 4.2.6 on 2023-10-26 13:54

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        (
            "wbwriter",
            "0001_initial_squashed_squashed_0008_alter_article_author_alter_article_feedback_contact_and_more",
        ),
    ]

    operations = [
        migrations.CreateModel(
            name="DependantArticle",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                (
                    "article",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="dependant_article_connections",
                        to="wbwriter.article",
                    ),
                ),
                (
                    "dependant_article",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="used_article_connections",
                        to="wbwriter.article",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
    ]



from wbcore.viewsets import ModelViewSet, RepresentationViewSet

from wbwriter.models import InEditorTemplate
from wbwriter.serializers import (
    InEditorTemplateModelSerializer,
    InEditorTemplateRepresentationSerializer,
)

from .display import InEditorTemplateDisplayConfig


class InEditorTemplateModelViewSet(ModelViewSet):
    ENDPOINT = "wbwriter:in-editor-template-list"
    IDENTIFIER = "wbwriter:in-editor-template"

    serializer_class = InEditorTemplateModelSerializer
    queryset = InEditorTemplate.objects.all()

    display_config_class = InEditorTemplateDisplayConfig

    search_fields = (
        "uuid",
        "title",
        "description",
    )
    filter_fields = {
        "uuid": ["icontains", "exact"],
        "title": ["icontains", "exact"],
        "description": [
            "icontains",
        ],
        "is_stand_alone_template": [
            "exact",
        ],
    }
    ordering_fields = (
        "uuid",
        "title",
        "is_stand_alone_template",
    )


class InEditorTemplateRepresentationViewSet(RepresentationViewSet):
    ENDPOINT = "wbwriter:in-editor-template-list"
    IDENTIFIER = "wbwriter:in-editor-template"

    serializer_class = InEditorTemplateRepresentationSerializer
    queryset = InEditorTemplate.objects.all()

    search_fields = (
        "uuid",
        "title",
        "description",
    )
    filter_fields = {
        "uuid": ["icontains", "exact"],
        "title": ["icontains", "exact"],
        "description": [
            "icontains",
        ],
        "is_stand_alone_template": [
            "exact",
        ],
    }
    ordering_fields = (
        "uuid",
        "title",
        "is_stand_alone_template",
    )



from .article import *
from .article_type import *
from .block import *
from .block_parameter import *
from .in_editor_template import *
from .meta_information import *
from .publication_models import *
from .style import *
from .template import *



import datetime
import json
import re
from contextlib import suppress
from copy import deepcopy
from datetime import date, timedelta

from celery import shared_task
from django import template
from django.contrib.auth import get_user_model
from django.contrib.contenttypes.fields import GenericRelation
from django.core.files.base import ContentFile
from django.db import models
from django.db.models import Count, Max, OuterRef, Subquery
from django.db.models.functions import Coalesce
from django.utils.translation import gettext_lazy
from django_fsm import FSMField, transition
from dynamic_preferences.registries import global_preferences_registry
from modeltrans.fields import TranslationField
from rest_framework.reverse import reverse
from slugify import slugify
from wbcore.contrib.directory.models import Person
from wbcore.contrib.documents.models import Document, DocumentType
from wbcore.contrib.i18n.translation import translate_string
from wbcore.contrib.icons import WBIcon
from wbcore.contrib.notifications.dispatch import send_notification
from wbcore.contrib.notifications.utils import create_notification_type
from wbcore.contrib.tags.models import TagModelMixin
from wbcore.enums import RequestType
from wbcore.markdown.template import resolve_content
from wbcore.metadata.configs.buttons import ActionButton, ButtonDefaultColor
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    create_simple_display,
)
from wbcore.models import WBModel
from wbcore.utils.models import CloneMixin
from weasyprint import CSS

from wbwriter.models.publication_models import Publication
from wbwriter.pdf_generator import PdfGenerator
from wbwriter.publication_parser import ParserValidationException
from wbwriter.typings import ArticleDTO

from .mixins import PublishableMixin


@shared_task
def generate_publications(article_id):
    with suppress(Article.DoesNotExist):
        article = Article.objects.get(id=article_id)
        if article.can_be_published():
            for parser in article.type.parsers.all():
                # Create the publication
                Publication.create_or_update_from_parser_and_object(parser, article)


def can_administrate_article(instance, user):
    """Allow only superusers and article admins to admin articles."""
    is_superuser = user.is_superuser
    is_article_admin = user.has_perm("wbwriter.administrate_article")

    return is_superuser or is_article_admin


def can_access_article(instance, user):
    """Allow access to superusers always, and to others when their position is appropriate."""
    return (
        can_administrate_article(instance, user)
        or user.is_superuser
        or user.profile.is_internal
        or is_reviewer(instance, user)
        or is_qa_reviewer(instance, user)
        or is_peer_reviewer(instance, user)
    )


def can_edit_article_author(instance, user) -> bool:
    """Allow the author and admins to edit the author."""
    if instance.author is None:
        return True

    return can_administrate_article(instance, user)


def can_edit_article_content(instance, user) -> bool:
    """Allow the appropriate role to edit content based on the state of the article."""
    if instance.author is None:
        return True

    if instance.status == Article.Status.DRAFT:
        return is_author(instance, user)

    if instance.status == Article.Status.FEEDBACK:
        return is_reviewer(instance, user)

    if instance.status == Article.Status.PEER_REVIEW:
        return is_peer_reviewer(instance, user)

    if instance.status == Article.Status.QA_REVIEW:
        return is_qa_reviewer(instance, user)

    return can_administrate_article(instance, user)


def can_edit_article_meta_data(instance, user) -> bool:
    """Allow the author to change the meta data of the article."""
    return (
        is_author(instance, user)
        or (is_reviewer(instance, user) and instance.status == Article.Status.FEEDBACK)
        or (is_peer_reviewer(instance, user) and instance.status == Article.Status.PEER_REVIEW)
        or (is_qa_reviewer(instance, user) and instance.status == Article.Status.QA_REVIEW)
    )


def can_edit_article_type(instance, user) -> bool:
    return is_author(instance, user) and instance.status == Article.Status.DRAFT


def can_request_peer_review(instance) -> bool:
    return not instance.peer_reviewer_approved


def can_request_qa_review(instance) -> bool:
    return instance.peer_reviewer_approved


def is_author(instance, user) -> bool:
    """Confirm user is the author of the instance."""
    return instance.author is None or instance.author == user.profile or can_administrate_article(instance, user)


def is_reviewer(instance, user) -> bool:
    """Confirm user is the reviewer of the instance."""
    return instance.reviewer == user.profile or can_administrate_article(instance, user)


def is_peer_reviewer(instance, user) -> bool:
    """Confirm user is ine of the peer reviewer of the instance's type."""
    if can_administrate_article(instance, user):
        return True
    return (
        instance.peer_reviewer is not None and instance.peer_reviewer.id == user.profile.id
    ) or instance.type.peer_reviewers.filter(id=user.profile.id).exists()


def is_qa_reviewer(instance, user) -> bool:
    """Confirm user is the quality assurance reviewer of the instance's type."""
    if can_administrate_article(instance, user):
        return True
    return (
        instance.qa_reviewer is not None and instance.qa_reviewer.id == user.profile.id
    ) or instance.type.qa_reviewers.filter(id=user.profile.id).exists()


class DependantArticle(WBModel):
    article = models.ForeignKey(
        to="wbwriter.Article",
        related_name="dependant_article_connections",
        on_delete=models.CASCADE,
    )

    dependant_article = models.ForeignKey(
        to="wbwriter.Article",
        related_name="used_article_connections",
        on_delete=models.PROTECT,
    )

    @classmethod
    def get_endpoint_basename(self):
        return "wbwriter:dependantarticle"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbwriter:dependantarticle-list"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{ article }} - {{ dependant_article }}"


def create_dependencies(article):
    if article.id:
        for dependency in re.findall("{% load_article.* ([0-9]*) %}", json.dumps(article.content)):
            DependantArticle.objects.get_or_create(article=article, dependant_article_id=dependency)


class Article(CloneMixin, TagModelMixin, PublishableMixin, WBModel):
    class Status(models.TextChoices):
        DRAFT = ("draft", "Draft")
        FEEDBACK = ("feedback", "Feedback")
        PEER_REVIEW = ("peer_review", "Peer Review")
        QA_REVIEW = ("qa_review", "QA Review")
        AUTHOR_APPROVAL = ("author_approval", "Author approval")
        APPROVED = ("approved", "Approved")
        PUBLISHED = ("published", "Published")

    class Meta:
        verbose_name = "Article"
        verbose_name_plural = "Articles"

        permissions = [
            ("administrate_article", "Can administrate Articles."),
        ]

        notification_types = [
            create_notification_type(
                "wbwriter.article.notify",
                "Article Notification",
                "Sends a notification when something happens in a relevant article.",
                True,
                True,
                False,
            ),
        ]

    """
    /////////////////////////////////////////////////////////////////
    ///                  FSM Transition Buttons                   ///
    /////////////////////////////////////////////////////////////////
    """
    fsm_base_button_parameters = {
        "method": RequestType.PATCH,
        "identifiers": ("wbwriter:article",),
        "description_fields": "<p>{{ title }}</p>",
    }

    request_feedback_button = ActionButton(
        method=RequestType.PATCH,
        identifiers=("wbwriter:article",),
        description_fields="<p>Select a contact to ask for feedback.</p>",
        key="requestfeedback",
        label="Request Feedback",
        action_label="Request Feedback",
        color=ButtonDefaultColor.PRIMARY,
        icon=WBIcon.FEEDBACK.icon,
        instance_display=create_simple_display([["feedback_contact"]]),
    )

    submit_feedback_button = ActionButton(
        method=RequestType.PATCH,
        identifiers=("wbwriter:article",),
        description_fields="<p>Are you sure you want to<br />submit your feedback?</p>",
        key="submitfeedback",
        label="Submit Feedback",
        action_label="Submit Feedback",
        color=ButtonDefaultColor.PRIMARY,
        icon=WBIcon.SEND.icon,
    )

    request_peer_review_button = ActionButton(
        method=RequestType.PATCH,
        identifiers=("wbwriter:article",),
        description_fields="<p>Are you sure you want to<br />request a peer review?</p>",
        key="requestpeerreview",
        label="Request Peer Review",
        action_label="Request Peer Review",
        color=ButtonDefaultColor.PRIMARY,
        icon=WBIcon.PEOPLE.icon,
    )

    request_qa_review_button = ActionButton(
        method=RequestType.PATCH,
        identifiers=("wbwriter:article",),
        description_fields="<p>Are you sure you want to<br />request a QA review?</p>",
        key="requestqareview",
        label="Request QA Review",
        action_label="Request QA Review",
        color=ButtonDefaultColor.PRIMARY,
        icon=WBIcon.PEOPLE.icon,
    )

    peer_approve_button = ActionButton(
        method=RequestType.PATCH,
        identifiers=("wbwriter:article",),
        description_fields="<p>Are you sure you want to<br />approve this draft?</p>",
        key="peerapprove",
        label="Approve",
        action_label="Approve",
        color=ButtonDefaultColor.SUCCESS,
        icon=WBIcon.CONFIRM.icon,
    )

    peer_reject_button = ActionButton(
        method=RequestType.PATCH,
        identifiers=("wbwriter:article",),
        description_fields="<p>Are you sure you want to<br />reject this draft?</p>",
        key="peerreject",
        label="Request Changes",
        action_label="Request Changes",
        color=ButtonDefaultColor.WARNING,
        icon=WBIcon.REJECT.icon,
    )

    qa_approve_draft_button = ActionButton(
        method=RequestType.PATCH,
        identifiers=("wbwriter:article",),
        description_fields="<p>Are you sure you want to<br />approve this draft?</p>",
        key="qaapprove",
        label="Approve",
        action_label="Approve",
        color=ButtonDefaultColor.SUCCESS,
        icon=WBIcon.APPROVE.icon,
    )

    qa_reject_draft_button = ActionButton(
        method=RequestType.PATCH,
        identifiers=("wbwriter:article",),
        description_fields="<p>Are you sure you want to<br />reject this draft?</p>",
        key="qareject",
        label="Request Changes",
        action_label="Request Changes",
        color=ButtonDefaultColor.WARNING,
        icon=WBIcon.DENY.icon,
    )

    author_approve_button = ActionButton(
        method=RequestType.PATCH,
        identifiers=("wbwriter:article",),
        description_fields="<p>Are you sure you want to<br />approve this article?</p>",
        key="authorapprove",
        label="Approve",
        action_label="Approve",
        color=ButtonDefaultColor.SUCCESS,
        icon=WBIcon.APPROVE.icon,
    )

    author_reject_button = ActionButton(
        method=RequestType.PATCH,
        identifiers=("wbwriter:article",),
        description_fields="<p>Are you sure you want to<br />reject this article?</p>",
        key="authorreject",
        label="Reject",
        action_label="Reject",
        color=ButtonDefaultColor.WARNING,
        icon=WBIcon.DENY.icon,
    )

    authors_revise_button = ActionButton(
        method=RequestType.PATCH,
        identifiers=("wbwriter:article",),
        description_fields="<p>Are you sure you want to<br />revise this article?</p>",
        key="authorrevise",
        label="Revise Article",
        action_label="Revise Article",
        color=ButtonDefaultColor.PRIMARY,
        icon=WBIcon.SYNCHRONIZE.icon,
    )

    qas_revise_button = ActionButton(
        method=RequestType.PATCH,
        identifiers=("wbwriter:article",),
        description_fields="<p>Are you sure you want to<br />revise this article's review?</p>",
        key="qarevise",
        label="Revise Review",
        action_label="Revise Review",
        color=ButtonDefaultColor.PRIMARY,
        icon=WBIcon.SYNCHRONIZE.icon,
    )

    publish_button = ActionButton(
        method=RequestType.PATCH,
        identifiers=("wbwriter:article",),
        description_fields="<p>Are you sure you want to<br />publish this article?</p>",
        key="publish",
        label="Publish",
        action_label="Publish",
        color=ButtonDefaultColor.PRIMARY,
        icon=WBIcon.DOCUMENT.icon,
    )

    unpublish_button = ActionButton(
        method=RequestType.PATCH,
        identifiers=("wbwriter:article",),
        description_fields="<p>Are you sure you want to<br />revise this article's publication?</p>",
        key="unpublish",
        label="Unpublish",
        action_label="Unpublish",
        color=ButtonDefaultColor.PRIMARY,
        icon=WBIcon.UNDO.icon,
    )

    used_article_connections: models.QuerySet[DependantArticle]
    dependant_article_connections: models.QuerySet[DependantArticle]

    """
    /////////////////////////////////////////////////////////////////
    ///                  /FSM Transition Buttons                  ///
    /////////////////////////////////////////////////////////////////
    """

    name = models.CharField(
        max_length=1024,
        unique=True,
        help_text="A unique name to reference this article.",
    )
    slug = models.CharField(max_length=1024, null=True, blank=True)
    title = models.CharField(
        max_length=1024,
        null=True,
        blank=True,
        help_text="The title of the article that is going to be used when imported into other articles."
        + " Defaults to the name of the article when not set.",
    )
    teaser_image = models.ImageField(blank=True, null=True, upload_to="writer/article/teasers")
    created = models.DateField(
        verbose_name="Creation Date",
        auto_now_add=True,
        help_text="The date on which this article has been created.",
    )
    modified = models.DateTimeField(
        verbose_name="Last modification date and time",
        auto_now=True,
        help_text="The last time this article has been edited.",
    )
    content = models.JSONField(
        verbose_name="Content",
        default=dict,
        blank=False,
        null=False,
    )

    i18n = TranslationField(fields=["title", "content", "slug"])

    def _translate_content(self, to_language: str) -> dict:
        new_content = deepcopy(self.content)
        for section_key in new_content.get("sections", {}):
            for content_key, content_value in new_content["sections"][section_key]["content"].items():
                new_content["sections"][section_key]["content"][content_key] = translate_string(
                    content_value, to_language
                )
        return new_content

    type = models.ForeignKey(
        "wbwriter.ArticleType",
        related_name="article",
        on_delete=models.PROTECT,
        blank=False,
        null=False,
    )

    publications = GenericRelation(Publication)

    """
    /////////////////////////////////////////////////////////////////
    ///                  Access relevant fields                   ///
    /////////////////////////////////////////////////////////////////
    TODO:
        - Protect the author field. Only the "admin role" can change this.
        - We should have an adjustable value for the minimum number of required peer reviews.
    """
    author = models.ForeignKey(
        "directory.Person",
        related_name="author_articles",
        blank=True,
        null=True,
        on_delete=models.PROTECT,
    )
    feedback_contact = models.ForeignKey(
        "directory.Person",
        related_name="feedback_contact_articles",
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
    )
    reviewer = models.ForeignKey(
        "directory.Person",
        related_name="review_articles",
        help_text="The contact that is currently working on feedback.",
        blank=True,
        null=True,
        on_delete=models.SET_NULL,  # TODO: We should transition back to DRAFT when the contact gets deleted.
    )

    peer_reviewer = models.ForeignKey(
        "directory.Person",
        related_name="peer_review_articles",
        help_text="The peer reviewer who reviewed this article.",
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
    )
    peer_reviewer_approved = models.BooleanField(default=False)
    qa_reviewer = models.ForeignKey(
        "directory.Person",
        related_name="qa_review_articles",
        help_text="The quality assurance (QA) reviewer who reviewed this article.",
        blank=True,
        null=True,
        on_delete=models.PROTECT,
    )

    is_private = models.BooleanField(
        blank=False,
        null=False,
        default=False,
        help_text="Signifies whether this article can be seen by all customers on the homepage.",
    )

    """
    /////////////////////////////////////////////////////////////////
    ///                  /Access relevant fields                  ///
    /////////////////////////////////////////////////////////////////
    """

    # QUESTION: What are we going to do with this when IETs become the norm?
    template = models.ForeignKey(
        "wbwriter.Template",
        related_name="articles",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
    )

    article_structure = models.JSONField(null=True, blank=True)

    def get_publication_metadata(self) -> dict[str, str]:
        """Returns the title, slug, author, and teasre image for the
        publication as a dictionary.
        """
        return {
            "title": self.title,
            "slug": slugify(self.title),
            "author": self.author,
        }

    def get_tag_detail_endpoint(self):
        return reverse("wbwriter:article-detail", [self.id])

    def get_tag_representation(self):
        return self.name

    """
    /////////////////////////////////////////////////////////////////
    ///                      FSM Transitions                      ///
    /////////////////////////////////////////////////////////////////
    """
    status = FSMField(choices=Status.choices, default=Status.DRAFT)

    @transition(
        status,
        source=[Status.DRAFT],
        target=Status.FEEDBACK,
        permission=is_author,
        custom={"_transition_button": request_feedback_button},
    )
    def requestfeedback(self, by=None):
        """Submit a draft to a reviewer (internal or external) that may give feedback."""
        self.reviewer = self.feedback_contact
        if user := getattr(self.reviewer, "user_account", None):
            send_notification(
                code="wbwriter.article.notify",
                title="Feedback requested",
                body=f"{by.profile} has requested feedback from you.",
                user=user,
                reverse_name="wbwriter:article-detail",
                reverse_args=[self.id],
            )

    @transition(
        status,
        source=[Status.FEEDBACK],
        target=Status.DRAFT,
        permission=is_reviewer,
        custom={"_transition_button": submit_feedback_button},
    )
    def submitfeedback(self, by=None):
        """Submit the feedback to the author."""
        if user := getattr(self.author, "user_account", None):
            send_notification(
                code="wbwriter.article.notify",
                title="Feedback received",
                body=f"{by.profile} has send your their feedback.",
                user=user,
                reverse_name="wbwriter:article-detail",
                reverse_args=[self.id],
            )

    @transition(
        status,
        source=[Status.DRAFT],
        target=Status.PEER_REVIEW,
        permission=is_author,
        custom={"_transition_button": request_peer_review_button},
        conditions=[can_request_peer_review],
    )
    def requestpeerreview(self, by=None):
        """Submit a draft to a peer reviewer that may approve or reject the draft."""
        if user := getattr(self.peer_reviewer, "user_account", None):
            send_notification(
                code="wbwriter.article.notify",
                title="Selected for peer review",
                body=f'You have been selected to review the draft of "{self.title}" from {self.author}.',
                user=user,
                reverse_name="wbwriter:article-detail",
                reverse_args=[self.id],
            )

    @transition(
        status,
        source=[Status.DRAFT],
        target=Status.QA_REVIEW,
        permission=is_author,
        custom={"_transition_button": request_qa_review_button},
        conditions=[can_request_qa_review],
    )
    def requestqareview(self, by=None):
        """Submit a draft to a QA reviewer that may approve or reject the draft."""
        if user := getattr(self.qa_reviewer, "user_account", None):
            send_notification(
                code="wbwriter.article.notify",
                title="QA review requested",
                body=f'{by.profile} has requested your review of "{self.title}".',
                user=user,
                reverse_name="wbwriter:article-detail",
                reverse_args=[self.id],
            )

    @transition(
        status,
        source=[Status.PEER_REVIEW],
        target=Status.QA_REVIEW,
        permission=is_peer_reviewer,
        custom={"_transition_button": peer_approve_button},
    )
    def peerapprove(self, by=None):
        """Approve the article and send it to the QA reviewer."""
        self.peer_reviewer_approved = True
        if user := getattr(self.author, "user_account", None):
            send_notification(
                code="wbwriter.article.notify",
                title="Peer approved your article",
                body=f'{by.profile} has approved your article "{self.title}" and send it to {self.qa_reviewer} for approval.',
                user=user,
                reverse_name="wbwriter:article-detail",
                reverse_args=[self.id],
            )
        if user := getattr(self.qa_reviewer, "user_account", None):
            send_notification(
                code="wbwriter.article.notify",
                title="Selected for QA review",
                body=f'{by.profile} has approved "{self.title}" from {self.author} and you have been selected to review it.',
                user=user,
                reverse_name="wbwriter:article-detail",
                reverse_args=[self.id],
            )

    @transition(
        status,
        source=[Status.PEER_REVIEW],
        target=Status.DRAFT,
        permission=is_peer_reviewer,
        custom={"_transition_button": peer_reject_button},
    )
    def peerreject(self, by=None):
        """Reject the article and send it back to the author."""
        self.peer_reviewer_approved = False
        if user := getattr(self.author, "user_account", None):
            send_notification(
                code="wbwriter.article.notify",
                title=f"{by.profile} rejected your article",
                body=f'{by.profile} has requested changes to your article "{self.title}" and send you their feedback.',
                user=user,
                reverse_name="wbwriter:article-detail",
                reverse_args=[self.id],
            )

    @transition(
        status,
        source=[Status.QA_REVIEW],
        target=Status.DRAFT,
        permission=is_qa_reviewer,
        custom={"_transition_button": qa_reject_draft_button},
    )
    def qareject(self, by=None):
        """Reject the article and return it to the peer reviewer."""
        if user := getattr(self.author, "user_account", None):
            send_notification(
                code="wbwriter.article.notify",
                title="Your article failed the QA review",
                body=f'{by.profile} has rejected "{self.title}".',
                user=user,
                reverse_name="wbwriter:article-detail",
                reverse_args=[self.id],
            )

    @transition(
        status,
        source=[Status.QA_REVIEW],
        target=Status.AUTHOR_APPROVAL,
        permission=is_qa_reviewer,
        custom={"_transition_button": qa_approve_draft_button},
    )
    def qaapprove(self, by=None):
        """Approve the article."""
        if user := getattr(self.author, "user_account", None):
            send_notification(
                code="wbwriter.article.notify",
                title=f"{by.profile} approved your article",
                body=f'{by.profile} has approved "{self.title}". Please, review.',
                user=user,
                reverse_name="wbwriter:article-detail",
                reverse_args=[self.id],
            )

    @transition(
        status,
        source=[Status.AUTHOR_APPROVAL],
        target=Status.QA_REVIEW,
        permission=is_author,
        custom={"_transition_button": author_reject_button},
    )
    def authorreject(self, by=None):
        """Reject the article and return it to the QA reviewer."""
        if user := getattr(self.qa_reviewer, "user_account", None):
            send_notification(
                code="wbwriter.article.notify",
                title=f"{by.profile} did not approve of your review",
                body=f'{by.profile} has rejected your review of "{self.title}".',
                user=user,
                reverse_name="wbwriter:article-detail",
                reverse_args=[self.id],
            )

    @transition(
        status,
        source=[Status.AUTHOR_APPROVAL],
        target=Status.APPROVED,
        permission=is_author,
        custom={"_transition_button": author_approve_button},
        # conditions=[can_qa_approve],
    )
    def authorapprove(self, by=None):
        """Approve the reviewed article."""
        if user := getattr(self.qa_reviewer, "user_account", None):
            send_notification(
                code="wbwriter.article.notify",
                title=f"{by.profile} approved your review",
                body=f'{by.profile} has approved "{self.title}". It is now in the "Approved" state.',
                user=user,
                reverse_name="wbwriter:article-detail",
                reverse_args=[self.id],
            )

    @transition(
        status,
        source=[Status.APPROVED],
        target=Status.PUBLISHED,
        permission=can_administrate_article,
        custom={"_transition_button": publish_button},
    )
    def publish(self, by=None):
        """Publish the article."""
        if self.publications.all().count() == 0 and (user := getattr(self.author, "user_account", None)):
            # Send a notification only on the first time the article has been published.
            # Re-publishing the article should not bother the author.
            send_notification(
                code="wbwriter.article.notify",
                title="Your article has been published",
                body=f'Your article "{self.title}" has been published.',
                user=user,
                reverse_name="wbwriter:article-detail",
                reverse_args=[self.id],
            )
        if self.type is not None:
            generate_publications.delay(self.id)

    def can_publish(self) -> dict[str, str]:
        """Allow only one-off and deep-dive articles to be published."""
        errors = dict()
        if self.status not in [self.Status.APPROVED, self.Status.PUBLISHED]:
            errors["status"] = [gettext_lazy("Status needs to be approved in order to allow publication")]
        if not self.type or self.type.slug not in [
            "one-off-article",
            "deep-dive-article",
            "mid-year-review",
            "year-s-favorites",
        ]:
            errors["type"] = [gettext_lazy("unvalid type for publication")]
        # We ensure the article's parser can be published
        if article_type := self.type:
            for parser in article_type.parsers.all():
                try:
                    parser.parser_class(
                        self._build_dto(),
                        date.today(),
                    ).is_valid()
                except ParserValidationException as e:
                    errors["non_field_errors"] = e.errors
                except ModuleNotFoundError:
                    errors["non_field_errors"] = [gettext_lazy("invalid parser")]
        return errors

    def can_be_published(self) -> bool:
        return not bool(self.can_publish())

    @transition(
        status,
        source=[Status.PUBLISHED],
        target=Status.APPROVED,
        permission=can_administrate_article,
        custom={"_transition_button": unpublish_button},
    )
    def unpublish(self, by=None):
        """Move the article back to the approved state."""
        pass

    # @transition(
    #     status,
    #     source=[Status.APPROVED, Status.PUBLISHED],
    #     target=Status.DRAFT,
    #     permission=can_revise,
    #     custom={"_transition_button": revise_button},
    # )
    # def revise(self, by=None):
    #     """Move the article back to the draft state."""
    #     self.peer_reviewer_approved = False
    #     self.peer_reviewer = None
    #     if self.type.slug == "ddq":
    #         self.author = None

    @transition(
        status,
        source=[Status.APPROVED],
        target=Status.DRAFT,
        permission=is_author,
        custom={"_transition_button": authors_revise_button},
    )
    def authorrevise(self, by=None):
        """Move the article back to the draft state."""
        pass

    @transition(
        status,
        source=[Status.APPROVED],
        target=Status.QA_REVIEW,
        permission=is_qa_reviewer,
        custom={"_transition_button": qas_revise_button},
    )
    def qarevise(self, by=None):
        """Move the article back to the qa_review state."""
        pass

    """
    /////////////////////////////////////////////////////////////////
    ///                     /FSM Transitions                      ///
    /////////////////////////////////////////////////////////////////
    """

    def __str__(self):
        return self.title

    def save(self, *args, **kwargs):
        self.slug = slugify(self.name)
        self.article_structure = self.get_sub_article_titles()
        if self.title in ("", None):
            self.title = self.name
        # Set the reviewers
        if not self.peer_reviewer:
            self.peer_reviewer = self.roll_peer()
        if not self.qa_reviewer:
            self.qa_reviewer = self.roll_qa()

        # if self.type and self.type.slug == "ddq":
        #     with suppress(Person.DoesNotExist):
        #         self.qa_reviewer = Person.objects.get(computed_str__icontains="Stefano Rodella")

        create_dependencies(self)
        super().save(*args, **kwargs)

    def _clone(self, user=None, **kwargs) -> "Article":
        from wbwriter.models import MetaInformationInstance

        object_copy = Article.objects.get(id=self.id)
        object_copy.id = None
        object_copy.name += "_clone"
        object_copy.title += " (Clone)"
        if user and (profile := user.profile):
            object_copy.author = profile
        object_copy.save()
        for meta_rel in self.meta_information.all():
            if not MetaInformationInstance.objects.filter(
                article=object_copy, meta_information=meta_rel.meta_information
            ).exists():
                meta_rel.article = object_copy
                meta_rel.id = None
                meta_rel.save()
        for dependant_article in self.dependant_article_connections.all():
            dependant_article.article = object_copy
            dependant_article.id = None
            dependant_article.save()
        for dependant_article_relationship in self.used_article_connections.all():
            dependant_article_relationship.dependant_article = object_copy
            dependant_article_relationship.id = None
            dependant_article_relationship.save()

        return object_copy

    @property
    def system_key(self):
        return f"article-{self.id}"

    def _build_dto(self, **kwargs):
        meta_informations = kwargs.get("meta_information", {})
        tags = kwargs.get("tags", [])
        if self.id:
            if hasattr(self, "meta_information"):
                meta_informations = list(
                    self.meta_information.values(
                        "meta_information__key",
                        "meta_information__name",
                        "meta_information__meta_information_type",
                        "meta_information__boolean_default",
                        "boolean_value",
                    )
                )
            if hasattr(self, "tags"):
                tags = list(self.tags.values_list("id", flat=True))
        return ArticleDTO(
            name=self.name,
            slug=getattr(self, "slug", slugify(self.name)),
            title=getattr(self, "title", self.name),
            teaser_image=getattr(self, "teaser_image", None),
            created=getattr(self, "created", datetime.datetime.now()),
            modified=getattr(self, "modified", datetime.datetime.now()),
            content=self.content,
            is_private=self.is_private,
            article_structure=getattr(
                self, "article_structure", dict()
            ),  # todo make the logic behind article_structure agnostic to the data model.
            status=self.status,
            meta_information=meta_informations,
            tags=tags,
        )

    def roll_peer(self):
        """Return a randomly chosen user profile from the list of peer
        reviewers of the type.

        Returns the author if no type has been assigned or if there are no peer
        reviewers listed on the assigned type.
        """
        if self.type:
            days_into_the_past = global_preferences_registry.manager()["wbwriter__reviewer_roll_days_range"]
            relevant_peers_query = self.type.article.filter(
                peer_reviewer=OuterRef("pk"), created__gte=date.today() - timedelta(days=days_into_the_past)
            ).values("peer_reviewer")
            num_articles = Subquery(relevant_peers_query.annotate(c=Count("*")).values("c")[:1])

            latest_article = Subquery(relevant_peers_query.annotate(c=Max("created")).values("c")[:1])

            peers = (
                self.type.peer_reviewers.exclude(id=self.author.id)
                .annotate(
                    num_articles=Coalesce(num_articles, 0),
                    latest_article=latest_article,
                )
                .order_by("num_articles", "latest_article")
            )
            if peers.exists():
                return peers.first()
        return self.author

    def roll_qa(self):
        """Return a randomly chosen user profile from the list of QA
        reviewers of the type.

        Returns the author if no type has been assigned or if there are no QA
        reviewers listed on the assigned type.
        """
        if self.type:
            days_into_the_past = global_preferences_registry.manager()["wbwriter__reviewer_roll_days_range"]
            relevant_qas_query = self.type.article.filter(
                qa_reviewer=OuterRef("pk"), created__gte=date.today() - timedelta(days=days_into_the_past)
            ).values("qa_reviewer")
            num_articles = Subquery(relevant_qas_query.annotate(c=Count("*")).values("c")[:1])

            latest_article = Subquery(relevant_qas_query.annotate(c=Max("created")).values("c")[:1])

            qas = (
                self.type.qa_reviewers.exclude(id=self.author.id)
                .annotate(
                    num_articles=Coalesce(num_articles, 0),
                    latest_article=latest_article,
                )
                .order_by("num_articles", "latest_article")
            )

            if qas.exists():
                return qas.first()
        return self.author

        qas = self.type.qa_reviewers.exclude(id=self.author.id) if self.author else self.type.qa_reviewers.all()
        if self.type:
            days_into_the_past = global_preferences_registry.manager()["wbwriter__reviewer_roll_days_range"]
            qa_reviewer_id = (
                Article.objects.filter(
                    qa_reviewer__in=qas, created__gte=date.today() - timedelta(days=days_into_the_past)
                )
                .values("qa_reviewer")
                .annotate(num_reviews=Count("*"), latest_review=Max("created"))
                .order_by("num_reviews", "latest_review")
                .values_list("qa_reviewer", flat=True)
            )
            if qa_reviewer_id.exists():
                return Person.objects.get(id=qa_reviewer_id.first())
        return self.author

    def reroll_peer(self):
        """Rerolls the peer reviewer and saves the newly rolled profile as the
        new peer reviewer.
        """
        self.peer_reviewer = self.roll_peer()

    def reroll_qa(self):
        """Rerolls the QA reviewer and saves the newly rolled profile as the
        new QA reviewer.
        """
        self.qa_reviewer = self.roll_qa()

    def reroll_peer_and_qa(self):
        """Rerolls the peer and QA reviewers and saves the newly rolled
        profiles as the new peer and QA reviewers.
        """
        self.peer_reviewer = self.roll_peer()
        self.qa_reviewer = self.roll_qa()

    def generate_pdf(self, user=None):
        empty_template = """<html>
    <head>
        <meta charset="UTF-8">
        <title>{{ self.title }}</title>
        <link rel="preconnect" href="https://fonts.googleapis.com">
        <link rel="preconnect" href="https://fonts.gstatic.com" crossorigin>
        <link
            href="https://fonts.googleapis.com/css2?family=Roboto:ital,wght@0,300;0,500;0,700;1,300&display=swap"
            rel="stylesheet"
        >
    </head>
    <body style="margin: 0">{{ content|safe }}</body>
</html>"""
        styles = []
        pdf_content = None
        if not self.content or "sections" not in self.content:
            html = template.Template(empty_template).render(
                {"content": "<h1>There seems to be no content to render.</h1>"}
            )
        else:
            content = self.content["sections"][self.content["sectionOrder"][0]]["content"]["content"]
            resolved_content = resolve_content(content, extensions=["sane_lists"], article=self)
            template_context = template.Context({"content": resolved_content})
            if self.type:
                if "ddq" in self.type.slug:
                    # TODO: Replace self.content with the raw_contents equivalent.

                    # If there is a template defined, then render that one and use its styles.
                    if self.template and self.template.template:
                        html = template.Template(self.template.template).render(template_context)

                        styles = [CSS(string=style.style) for style in self.template.styles.all()]
                        if header := self.template.header_template:
                            styles += [CSS(string=style.style) for style in header.styles.all()]
                        if footer := self.template.footer_template:
                            styles += [CSS(string=style.style) for style in footer.styles.all()]

                        pdf_content = PdfGenerator(
                            main_html=html,
                            header_html=self.template.header_template.template,
                            footer_html=self.template.footer_template.template,
                            custom_css=styles,
                            side_margin=self.template.side_margin,
                            extra_vertical_margin=self.template.extra_vertical_margin,
                        ).render_pdf()

                elif (
                    "one-off" in self.type.slug
                    or "deep-dive" in self.type.slug
                    or "mid-year" in self.type.slug
                    or "year-s-favorites" in self.type.slug
                ):
                    parsers = self.type.parsers.all()
                    if parsers.exists():
                        parser = parsers.first().parser_class(self._build_dto(), datetime.date.today())
                        if parser.is_valid(raise_exception=False):
                            pdf_content = parser.get_file()
                        else:
                            template_context["content"] = f"""
                            <p>To generate the PDF, please correct the following errors:</p>
                            <ul>
                            {"".join([f"<li>{error}</li>" for error in parser.errors])}
                            </ul>
                            """
            html = template.Template(empty_template).render(template_context)
        if not pdf_content:
            pdf_content = PdfGenerator(
                main_html=html,
                custom_css=styles,
                side_margin=0,
                extra_vertical_margin=0,
            ).render_pdf()
        document_type, created = DocumentType.objects.get_or_create(name="article")
        filename = f"{self.slug}.pdf"
        content_file = ContentFile(pdf_content, name=filename)
        document, created = Document.objects.update_or_create(
            system_created=True,
            system_key=f"article-{self.id}",
            defaults={
                "document_type": document_type,
                "file": content_file,
                "name": filename,
                "permission_type": Document.PermissionType.INTERNAL,
            },
        )
        document.link(self)

        if not user:
            user = getattr(self.author, "user_account", None)
        if user:
            send_notification(
                code="wbwriter.article.notify",
                title="Your article has been generated",
                body=f'Your article "{self.title}" has been generated.',
                user=user,
                reverse_name="wbwriter:article-detail",
                reverse_args=[self.id],
            )

    def re_find_articles(self) -> list:
        """Search in `content` for load_article template tags using regex and return the match."""
        # TODO: Fix the regex below. I belive that this regex doesn't represent
        #       the full range of possible load_article tremplate tags.
        load_article_regex = "{%[ ]*load_article[ ]+([0-9]*)[ ]*[A-Za-z]*[ ]+(False True)?[ ]*%}"

        if not self.content or "sectionOrder" not in self.content:
            return []

        result = []
        for section_id in self.content["sectionOrder"]:
            for content in self.content["sections"][section_id]["content"].values():
                result.extend(re.findall(load_article_regex, content))

        return result

    def get_article_structure(self, level=0, enumerator=None):
        articles = {
            "title": self.title,
            "anchor": slugify(self.name),
            "enumerator": enumerator,
            "level": level,
            "articles": list(),
        }

        for index, (sub_article_id, _) in enumerate(self.re_find_articles(), start=1):
            _enumerator = f"{enumerator or ''}{index}."
            articles["articles"].append(
                Article.objects.get(id=sub_article_id).get_article_structure(level + 1, _enumerator)
            )

        return articles

    def get_sub_article_titles(self, level=0, enumerator=None):
        mapping = dict()

        for index, (sub_article_id, _) in enumerate(self.re_find_articles(), start=1):
            _enumerator = f"{enumerator or ''}{index}."
            mapping[sub_article_id] = {"enumerator": _enumerator, "level": level}
            _mapping = Article.objects.get(id=sub_article_id).get_sub_article_titles(level + 1, _enumerator)
            mapping.update(**_mapping)
        return mapping

    @classmethod
    def get_endpoint_basename(self):
        return "wbwriter:article"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbwriter:article-list"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{ title }}"


@shared_task
def generate_pdf_as_task(article_id, user_id=None):
    article = Article.objects.get(id=article_id)
    user = get_user_model().objects.get(id=user_id) if user_id else None
    article.generate_pdf(user=user)
